
/* Common Open Systems Environment, Common Graphics Environment	 */		
/* Extension to	PEX 5.1	version	2.  File date: February	10, 1994 */		

/******************************************************************************/
/*  (c)	Copyright Hewlett-Packard Company, 1993,  Fort Collins,	Colorado      */
/*									      */
/*  (C)	COPYRIGHT International	Business Machines Corp.	1993		      */
/*									      */
/*   Copyright (c) 1993	by Sun Microsystems, Inc.			      */
/*									      */
/*			      All Rights Reserved			      */
/*									      */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for any purpose and without fee is hereby granted,	      */
/*  provided that the above copyright notices appear in	all copies and that   */
/*  both the copyright notices and this	permission notice appear in	      */
/*  supporting documentation, and that the names of Hewlett-Packard,	      */
/*  International Business Machines, or	Sun Microsystems not be	used in	      */
/*  advertising	or publicity pertaining	to distribution	of the software	      */
/*  without specific, written prior permission.				      */
/*									      */
/*  HEWLETT-PACKARD MAKES NO WARRANTY OF ANY KIND WITH REGARD TO THIS	      */
/*  SOFTWARE, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF	      */
/*  MERCHANTABILITY AND	FITNESS	FOR A PARTICULAR PURPOSE.  Hewlett-Packard    */
/*  shall not be liable	for errors contained herein or direct, indirect,      */
/*  special, incidental	or consequential damages in connection with the	      */
/*  furnishing,	performance or use of this software.			      */
/*									      */
/*  INTERNATIONAL BUSINESS MACHINES CORPORATION	PROVIDES THIS MATERIAL	      */
/*  "AS	IS" WITHOUT WARRANTY OF	ANY KIND, EITHER EXPRESSED OR IMPLIED,	      */
/*  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY  */
/*  AND	FITNESS	FOR A PARTICULAR PURPOSE.				      */
/*									      */
/*  SUN	MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES WITH REGARD	TO THIS	      */
/*  SOFTWARE, INCLUDING	ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,*/
/*  IN NO EVENT	SHALL SUN MICROSYSTEMS,	INC. BE	LIABLE FOR ANY SPECIAL,	      */
/*  INDIRECT OR	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER	RESULTING     */
/*  FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,      */
/*  NEGLIGENCE OR OTHER	TORTIOUS ACTION, ARISING OUT OF	OR IN CONNECTION      */
/*  WITH THE USE OR PERFORMANCE	OF THIS	SOFTWARE.			      */
/*									      */
/******************************************************************************/


 /* =======================================================================
  * NOTE NOTE NOTE  Insert this	section
  *		    at the end of pl_lut.c file
  * ====================================================================='=
  */

/***** Start of	Ext section ************/

#include "PEXExtlib.h"
#include "PEXExtlibint.h"

static PEXPointer _PEXExtRepackLUTEntries();


PEXPointer
PEXExtGetTableEntry (display, lut, index, valueType,
    statusReturn, table_type_return)

INPUT Display		*display;
INPUT PEXLookupTable	lut;
INPUT unsigned int	index;
INPUT int		valueType;
OUTPUT int		*statusReturn;
OUTPUT int		*table_type_return;

{
    register pexExtGetTableEntryReq	*req;
    char				*pBuf;
    pexExtGetTableEntryReply		*rep;
    int					fpConvert;
    int					fpFormat;
    int					escapeSize = 0;
    char				*escapeData;
    unsigned long			escapeOutDataSize;
    PEXPointer				entryReturn;

    /*
     * Fill in the escape record.
     */

    escapeSize = SIZEOF	(pexExtGetTableEntryReq);

    escapeData = PEXAllocBuf (escapeSize);

    fpFormat = PEXGetProtocolFloatFormat (display);
    fpConvert =	(fpFormat != NATIVE_FP_FORMAT);

    req	= (pexExtGetTableEntryReq *) escapeData;
    req->fpFormat = fpFormat;
    req->valueType = valueType;
    req->lut = lut;
    req->index = index;

    /*
     * Generate	the escape.
     */

    rep	= (pexExtGetTableEntryReply *) PEXEscapeWithReply (display,
	PEXExtEscapeOpcodeGetTableEntry, escapeSize, escapeData,
	&escapeOutDataSize);

    PEXFreeBuf (escapeData);

    if (rep == NULL) return (NULL);	    /* return an error */

    *statusReturn = rep->status;
    *table_type_return = (PEXEnumTypeIndex) rep->tableType;

    /*
     * Start the pBuf after the	pad bytes in the reply data
     */

    pBuf = (char *) rep	+ SIZEOF (pexExtGetTableEntryReply);

    /*
     * Repack the entries into a buffer	allocated for the application.
     */

    entryReturn	= _PEXExtRepackLUTEntries (pBuf, 1,
	(int) rep->tableType, fpConvert, fpFormat);

    PEXFreeBuf (rep);

    return (entryReturn);
}


Status
PEXExtGetTableEntries (display,	lut, start, count, valueType,
    table_type_return, entriesReturn)

INPUT Display		*display;
INPUT PEXLookupTable	lut;
INPUT unsigned int	start;
INPUT unsigned int	count;
INPUT int		valueType;
OUTPUT int		*table_type_return;
OUTPUT PEXPointer	*entriesReturn;

{
    register pexExtGetTableEntriesReq	*req;
    char				*pBuf;
    pexExtGetTableEntriesReply		*rep;
    int					fpConvert;
    int					fpFormat;
    int					escapeSize = 0;
    char				*escapeData;
    unsigned long			escapeOutDataSize;


    /*
     * Fill in the escape record.
     */

    escapeSize = SIZEOF	(pexExtGetTableEntriesReq);

    escapeData = PEXAllocBuf (escapeSize);

    fpFormat = PEXGetProtocolFloatFormat (display);
    fpConvert =	(fpFormat != NATIVE_FP_FORMAT);

    req	= (pexExtGetTableEntriesReq *) escapeData;
    req->fpFormat = fpFormat;
    req->valueType = valueType;
    req->lut = lut;
    req->start = start;
    req->count = count;

    /*
     * Generate	the escape.
     */

    rep	= (pexExtGetTableEntriesReply *) PEXEscapeWithReply (display,
	PEXExtEscapeOpcodeGetTableEntries, escapeSize, escapeData,
	&escapeOutDataSize);

    PEXFreeBuf (escapeData);
    if (rep == NULL)
    {
	*entriesReturn = NULL;
	return (0);	    /* return an error */
    }

    *table_type_return = (PEXEnumTypeIndex) rep->tableType;

    /*
     * Start the pBuf after the	pad bytes in the reply data
     */

    pBuf = (char *) rep	+ SIZEOF (pexExtGetTableEntriesReply);

    /*
     * Repack the entries into a buffer	allocated for the application.
     */

    *entriesReturn = _PEXExtRepackLUTEntries (pBuf, (int) rep->numEntries,
	(int) rep->tableType, fpConvert, fpFormat);

    PEXFreeBuf (rep);

    return (1);
}


void
PEXExtSetTableEntries (display,	lut, start, count, type, entries)

INPUT Display		*display;
INPUT PEXLookupTable	lut;
INPUT unsigned int	start;
INPUT unsigned int	count;
INPUT int		type;
INPUT PEXPointer	entries;

{
    register pexExtSetTableEntriesReq	*req;
    char				*pBuf;
    char				*scratch;
    int					fpConvert;
    int					fpFormat;
    int					size, i;
    int					escapeSize = 0;
    char				*escapeData;
    unsigned long			escapeOutDataSize;


    /*
     * Fill in the escape record (begin	with LUT entries)
     */

    fpFormat = PEXGetProtocolFloatFormat (display);
    fpConvert =	(fpFormat != NATIVE_FP_FORMAT);

    switch (type)
    {
    case PEXLUTLineBundle:
    {
	PEXLineBundleEntry *src	= (PEXLineBundleEntry *) entries;
	pexLineBundleEntry *dst;
	
	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXLineBundleEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexLineBundleEntry, scratch,	dst);

	    dst->lineType = src->type;
	    dst->polylineInterp	= src->interp_method;
	    dst->curveApprox_method = src->curve_approx.method;
	    dst->lineColorType = src->color.type;

	    if (fpConvert)
	    {
		FP_CONVERT_HTON	(src->curve_approx.tolerance,
		    dst->curveApprox_tolerance,	fpFormat);
		FP_CONVERT_HTON	(src->width, dst->lineWidth, fpFormat);
	    }
	    else
	    {
		dst->curveApprox_tolerance = src->curve_approx.tolerance;
		dst->lineWidth = src->width;
	    }

	    END_LUTENTRY_HEADER	(pexLineBundleEntry, scratch, dst);

	    STORE_COLOR_VAL (src->color.type, src->color.value,	scratch,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTMarkerBundle:
    {
	PEXMarkerBundleEntry *src = (PEXMarkerBundleEntry *) entries;
	pexMarkerBundleEntry *dst;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXMarkerBundleEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexMarkerBundleEntry, scratch, dst);

	    dst->markerType = src->type;
	    dst->markerColorType = src->color.type;

	    if (fpConvert)
	    {
		FP_CONVERT_HTON	(src->scale, dst->markerScale, fpFormat);
	    }
	    else
		dst->markerScale = src->scale;
	    
	    END_LUTENTRY_HEADER	(pexMarkerBundleEntry, scratch,	dst);

	    STORE_COLOR_VAL (src->color.type, src->color.value,	scratch,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTTextBundle:
    {
	PEXTextBundleEntry *src	= (PEXTextBundleEntry *) entries;
	pexTextBundleEntry *dst;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXTextBundleEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexTextBundleEntry, scratch,	dst);

	    dst->textFontIndex = src->font_index;
	    dst->textPrecision = src->precision;
	    dst->textColorType = src->color.type;

	    if (fpConvert)
	    {
		FP_CONVERT_HTON	(src->char_expansion,
		    dst->charExpansion,	fpFormat);
		FP_CONVERT_HTON	(src->char_spacing,
		    dst->charSpacing, fpFormat);
	    }
	    else
	    {
		dst->charExpansion = src->char_expansion;
		dst->charSpacing = src->char_spacing;
	    }

	    END_LUTENTRY_HEADER	(pexTextBundleEntry, scratch, dst);

	    STORE_COLOR_VAL (src->color.type, src->color.value,	scratch,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTInteriorBundle:
    {
	PEXInteriorBundleEntry *src = (PEXInteriorBundleEntry *) entries;
	pexInteriorBundleEntry *dst;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXInteriorBundleEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexInteriorBundleEntry, scratch, dst);

	    dst->interiorStyle = src->style;
	    dst->interiorStyleIndex = src->style_index;
	    dst->reflectionModel = src->reflection_model;
	    dst->surfaceInterp = src->interp_method;
	    dst->bfInteriorStyle = src->bf_style;
	    dst->bfInteriorStyleIndex =	src->bf_style_index;
	    dst->bfReflectionModel = src->bf_reflection_model;
	    dst->bfSurfaceInterp = src->bf_interp_method;
	    dst->surfaceApprox_method =	src->surface_approx.method;

	    if (fpConvert)
	    {
		FP_CONVERT_HTON	(src->surface_approx.u_tolerance,
		    dst->surfaceApproxuTolerance, fpFormat);
		FP_CONVERT_HTON	(src->surface_approx.v_tolerance,
		    dst->surfaceApproxvTolerance, fpFormat);
	    }
	    else
	    {
		dst->surfaceApproxuTolerance = src->surface_approx.u_tolerance;
		dst->surfaceApproxvTolerance = src->surface_approx.v_tolerance;
	    }

	    END_LUTENTRY_HEADER	(pexInteriorBundleEntry, scratch, dst);

	    /* copy surfaceColor */

	    STORE_COLOR_SPEC (src->color, scratch, fpConvert, fpFormat);

	    /* copy reflectionAttr */

	    STORE_REFLECTION_ATTR (src->reflection_attr, scratch,
		fpConvert, fpFormat);

	    /* copy bfSurfaceColor */

	    STORE_COLOR_SPEC (src->bf_color, scratch, fpConvert, fpFormat);

	    /* copy bfReflectionAttr */

	    STORE_REFLECTION_ATTR (src->bf_reflection_attr, scratch,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTEdgeBundle:
    {
	PEXEdgeBundleEntry *src	= (PEXEdgeBundleEntry *) entries;
	pexEdgeBundleEntry *dst;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXEdgeBundleEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexEdgeBundleEntry, scratch,	dst);

	    dst->edges = src->edge_flag;
	    dst->edgeType = src->type;
	    dst->edgeColorType = src->color.type;

	    if (fpConvert)
	    {		
		FP_CONVERT_HTON	(src->width, dst->edgeWidth, fpFormat);
	    }
	    else
		dst->edgeWidth = src->width;

	    END_LUTENTRY_HEADER	(pexEdgeBundleEntry, scratch, dst);

	    STORE_COLOR_VAL (src->color.type, src->color.value,	scratch,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTPattern:
    {
	PEXPatternEntry	*src = (PEXPatternEntry	*) entries;
	pexPatternEntry	*dst;

	size = count * sizeof (PEXPatternEntry);
	for (i = 0; i <	count; i++)
	    size += (src[i].col_count *	src[i].row_count * sizeof (PEXColor));

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) + size);
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexPatternEntry, scratch, dst);

	    dst->colorType = src->color_type;
	    dst->numx =	src->col_count;
	    dst->numy =	src->row_count;

	    END_LUTENTRY_HEADER	(pexPatternEntry, scratch, dst);

	    STORE_LISTOF_COLOR_VAL (src->col_count * src->row_count,
		src->color_type, src->colors, scratch, fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTTextFont:
    {
	PEXTextFontEntry *src =	(PEXTextFontEntry *) entries;
	pexTextFontEntry *dst;

	size = count * sizeof (PEXTextFontEntry);
	for (i = 0; i <	count; i++)
	    size += src[i].count * sizeof (PEXFont);

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) + size);
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexTextFontEntry, scratch, dst);

	    dst->numFonts = src->count;

	    END_LUTENTRY_HEADER	(pexTextFontEntry, scratch, dst);

	    STORE_LISTOF_CARD32	(src->count, src->fonts, scratch);
	}
	break;
    }

    case PEXLUTColor:
    {
	PEXColorEntry	  *src = (PEXColorEntry	*) entries;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXColorEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	STORE_LISTOF_COLOR_SPEC	(count,	src, scratch, fpConvert, fpFormat);
	break;
    }

    case PEXLUTView:
    {
	PEXViewEntry *src = (PEXViewEntry *) entries;
	pexViewEntry *dst;
	char	     *tPtr;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXViewEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexViewEntry, scratch, dst);

	    dst->clipFlags = src->clip_flags;

	    if (fpConvert)
	    {
		FP_CONVERT_HTON	(src->clip_limits.min.x,
		    dst->clipLimits_xmin, fpFormat);
		FP_CONVERT_HTON	(src->clip_limits.min.y,
		    dst->clipLimits_ymin, fpFormat);
		FP_CONVERT_HTON	(src->clip_limits.min.z,
		    dst->clipLimits_zmin, fpFormat);
		FP_CONVERT_HTON	(src->clip_limits.max.x,
		    dst->clipLimits_xmax, fpFormat);
		FP_CONVERT_HTON	(src->clip_limits.max.y,
		    dst->clipLimits_ymax, fpFormat);
		FP_CONVERT_HTON	(src->clip_limits.max.z,
		    dst->clipLimits_zmax, fpFormat);
	    }
	    else
	    {
		dst->clipLimits_xmin = src->clip_limits.min.x;
		dst->clipLimits_ymin = src->clip_limits.min.y;
		dst->clipLimits_zmin = src->clip_limits.min.z;
		dst->clipLimits_xmax = src->clip_limits.max.x;
		dst->clipLimits_ymax = src->clip_limits.max.y;
		dst->clipLimits_zmax = src->clip_limits.max.z;
	    }

	    tPtr = (char *) dst->orientation;
	    STORE_LISTOF_FLOAT32 (16, src->orientation,	tPtr,
		fpConvert, fpFormat);

	    tPtr = (char *) dst->mapping;
	    STORE_LISTOF_FLOAT32 (16, src->mapping, tPtr,
		fpConvert, fpFormat);
	    
	    END_LUTENTRY_HEADER	(pexViewEntry, scratch,	dst);
	}
	break;
    }

    case PEXLUTLight:
    {
	PEXLightEntry *src = (PEXLightEntry *) entries;
	pexLightEntry *dst;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXLightEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexLightEntry, scratch, dst);

	    dst->lightType = src->type;
	    dst->lightColorType	= src->color.type;

	    if (fpConvert)
	    {
		FP_CONVERT_HTON	(src->direction.x, dst->direction_x, fpFormat);
		FP_CONVERT_HTON	(src->direction.y, dst->direction_y, fpFormat);
		FP_CONVERT_HTON	(src->direction.z, dst->direction_z, fpFormat);

		FP_CONVERT_HTON	(src->point.x, dst->point_x, fpFormat);
		FP_CONVERT_HTON	(src->point.y, dst->point_y, fpFormat);
		FP_CONVERT_HTON	(src->point.z, dst->point_z, fpFormat);

		FP_CONVERT_HTON	(src->concentration,
		    dst->concentration,	fpFormat);
		FP_CONVERT_HTON	(src->spread_angle,
		    dst->spreadAngle, fpFormat);
		FP_CONVERT_HTON	(src->attenuation1,
		    dst->attenuation1, fpFormat);
		FP_CONVERT_HTON	(src->attenuation2,
		    dst->attenuation2, fpFormat);
	    }
	    else
	    {
		dst->direction_x = src->direction.x;
		dst->direction_y = src->direction.y;
		dst->direction_z = src->direction.z;

		dst->point_x = src->point.x;
		dst->point_y = src->point.y;
		dst->point_z = src->point.z;

		dst->concentration = src->concentration;
		dst->spreadAngle = src->spread_angle;
		dst->attenuation1 = src->attenuation1;
		dst->attenuation2 = src->attenuation2;
	    }

	    END_LUTENTRY_HEADER	(pexLightEntry,	scratch, dst);

	    STORE_COLOR_VAL (src->color.type, src->color.value,	scratch,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTDepthCue:
    {
	PEXDepthCueEntry *src =	(PEXDepthCueEntry *) entries;
	pexDepthCueEntry *dst;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXDepthCueEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexDepthCueEntry, scratch, dst);

	    dst->mode =	src->mode;
	    dst->depthCueColorType = src->color.type;

	    if (fpConvert)
	    {
		FP_CONVERT_HTON	(src->front_plane, dst->frontPlane, fpFormat);
		FP_CONVERT_HTON	(src->back_plane, dst->backPlane, fpFormat);

		FP_CONVERT_HTON	(src->front_scaling,
		    dst->frontScaling, fpFormat);
		FP_CONVERT_HTON	(src->back_scaling,
		    dst->backScaling, fpFormat);
	    }
	    else
	    {
		dst->frontPlane	= src->front_plane;
		dst->backPlane = src->back_plane;

		dst->frontScaling = src->front_scaling;
		dst->backScaling = src->back_scaling;
	    }
	     
	    END_LUTENTRY_HEADER	(pexDepthCueEntry, scratch, dst);

	    STORE_COLOR_VAL (src->color.type, src->color.value,	scratch,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTColorApprox:
    {
	PEXColorApproxEntry *src = (PEXColorApproxEntry	*) entries;
	pexColorApproxEntry *dst;

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXColorApproxEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexColorApproxEntry,	scratch, dst);

	    dst->approxType = src->type;
	    dst->approxModel = src->model;
	    dst->max1 =	src->max1;
	    dst->max2 =	src->max2;
	    dst->max3 =	src->max3;
	    dst->dither	= src->dither;
	    dst->mult1 = src->mult1;
	    dst->mult2 = src->mult2;
	    dst->mult3 = src->mult3;
	    dst->basePixel = src->base_pixel;

	    if (fpConvert)
	    {
		FP_CONVERT_HTON	(src->weight1, dst->weight1, fpFormat);
		FP_CONVERT_HTON	(src->weight2, dst->weight2, fpFormat);
		FP_CONVERT_HTON	(src->weight3, dst->weight3, fpFormat);
	    }
	    else
	    {
		dst->weight1 = src->weight1;
		dst->weight2 = src->weight2;
		dst->weight3 = src->weight3;
	    }

	    END_LUTENTRY_HEADER	(pexColorApproxEntry, scratch, dst);
	}
	break;
    }

    case PEXExtLUTTMBinding:	
    {
	PEXExtTMBindingEntry *src = (PEXExtTMBindingEntry *) entries;
	pexExtTMBindingEntry *dst;
	
	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXExtTMBindingEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexExtTMBindingEntry, scratch, dst);

	    dst->tmDescription = src->tm_description_id;
	    dst->coordinateLutIndex = src->coord_source_index;
	    dst->compositionLutIndex = src->composition_index;
	    dst->samplingLutIndex = src->sampling_index;     

	    END_LUTENTRY_HEADER	(pexExtTMBindingEntry, scratch,	dst);

	}
	break;
    }

    case PEXExtLUTTMCoordSource:
    {
	PEXExtTMCoordSourceEntry *src =	(PEXExtTMCoordSourceEntry *) entries;
	pexExtTMCoordSourceEntry *dst;
	char	     *tPtr;
	
	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXExtTMCoordSourceEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexExtTMCoordSourceEntry, scratch, dst);

	    dst->tmSource = src->tm_source; 
	    dst->fpDataIndex = src->fp_data_index;	    

	    tPtr = (char *) dst->orientation;
	    STORE_LISTOF_FLOAT32 (16, src->orientation,	tPtr,
		fpConvert, fpFormat);

	    END_LUTENTRY_HEADER	(pexExtTMCoordSourceEntry, scratch, dst);
	}
	break;
    }

    case PEXExtLUTTMComposition:
    {
	PEXExtTMCompositionEntry *src =	(PEXExtTMCompositionEntry *) entries;
	pexExtTMCompositionEntry *dst;
	char	     *tPtr;
	
	size = count * sizeof (PEXExtTMCompositionEntry);
	for (i = 0; i <	count; i++)
	    size += (src[i].method == PEXExtTMCompositeReplace ||  
		     src[i].method == PEXExtTMCompositeModulate	|| 
		     src[i].method == PEXExtTMCompositeDecal	|| 
		     src[i].method == PEXExtTMCompositeBlendEnvColor ||
		     src[i].method == PEXExtTMCompositeDecalBackground ||
		     src[i].method == PEXExtTMCompositeReplaceBlendedColors ||
		     src[i].method == PEXExtTMCompositeReplaceBlendedColors)
		     ? 0 : src[i].data.imp_dep.length;	 

	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) + size);
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexExtTMCompositionEntry, scratch, dst);

	    dst->method	= src->method; 

	    END_LUTENTRY_HEADER	(pexExtTMCompositionEntry, scratch, dst);

	    tPtr = scratch;   /* keep track of beginning composite method data*/
 
	    switch (src->method)
	    {
	    case PEXExtTMCompositeReplace:   
	    case PEXExtTMCompositeModulate:  
	    case PEXExtTMCompositeDecal:	 

		break;

	    case PEXExtTMCompositeBlendEnvColor:
	    { 
		STORE_CARD16 (src->data.blend_env.channel_selector, scratch); 
		scratch	+= 2;
		STORE_COLOR_SPEC (src->data.blend_env.color, scratch, 
		    fpConvert, fpFormat);
		break;
	    }

	    case PEXExtTMCompositeDecalBackground:
	    { 
		STORE_COLOR_SPEC (src->data.decal_bkgd_color, scratch,
		    fpConvert, fpFormat);
		break;
	    }

	    case PEXExtTMCompositeReplaceBlendedColors:
	    { 
		STORE_CARD16 (src->data.blend_repl.channel_selector, scratch); 
		scratch	+= 2;
		STORE_COLOR_SPEC (src->data.blend_repl.color1, scratch,	
		    fpConvert, fpFormat);
		STORE_COLOR_SPEC (src->data.blend_repl.color2, scratch,	
		    fpConvert, fpFormat);
		break;
	    }

	    default:					 
	    { 
		COPY_AREA ((char *) src->data.imp_dep.imp_dep, scratch,	
		    src->data.imp_dep.length); 
		scratch	+= src->data.imp_dep.length;
		break;
	    }
	    }
	     
	    dst->lengthCompositeData = scratch - tPtr;		    
	}
	break;
    }

    case PEXExtLUTTMSampling:
    {
	PEXExtTMSamplingEntry *src = (PEXExtTMSamplingEntry *) entries;
	pexExtTMSamplingEntry *dst;
	
	escapeData = PEXAllocBuf( SIZEOF (pexExtSetTableEntriesReq) +
	    count * sizeof (PEXExtTMSamplingEntry));
	scratch	= escapeData + SIZEOF (pexExtSetTableEntriesReq);

	for (i = 0; i <	count; i++, src++)
	{
	    BEGIN_LUTENTRY_HEADER (pexExtTMSamplingEntry, scratch, dst);

	    dst->minification =	src->minification_method; 
	    dst->magnification = src->magnification_method; 
	    dst->t0Boundary = src->t0_boundary_condition; 
	    dst->t1Boundary = src->t1_boundary_condition; 
	    dst->t2Boundary = src->t2_boundary_condition; 
	    dst->clampColorSource = src->boundary_clamp_color_source; 

	    END_LUTENTRY_HEADER	(pexExtTMSamplingEntry,	scratch, dst);

	    STORE_COLOR_SPEC (src->clamp_color,	scratch, fpConvert, fpFormat);

	    STORE_FLOAT32 (src->depth_sampling_bias_hint, scratch,
		fpConvert, fpFormat);

	    STORE_FLOAT32 (src->t0_frequency_hint, scratch, 
		fpConvert, fpFormat);
	    STORE_FLOAT32 (src->t1_frequency_hint, scratch, 
		fpConvert, fpFormat);
	    STORE_FLOAT32 (src->t2_frequency_hint, scratch, 
		fpConvert, fpFormat);
	}
	break;
    }
    }

    /*
     * Fill in the escape record's header information.
     */

    escapeSize = scratch - escapeData;		      

    req	= (pexExtSetTableEntriesReq *) escapeData;
    req->fpFormat = fpFormat;
    req->lut = lut;
    req->start = start;
    req->count = count;

    /*
     * Generate	the escape.
     */

    PEXEscape (display,	PEXExtEscapeOpcodeSetTableEntries,
	escapeSize, escapeData);

    PEXFreeBuf (escapeData);
}


/*
 * This	routine	repacks	the lut	entries	returned by PEXExtGetTableEntry,
 * and PEXExtGetTableEntries.  This is mostly to
 * change the colors from fixed	size to	the PEXColor union format,
 * as well as to do floating point conversion.
*/

static PEXPointer
_PEXExtRepackLUTEntries	(pBuf, numEntries, tableType, fpConvert, fpFormat)

INPUT char	*pBuf;
INPUT int	numEntries;
INPUT int	tableType;
INPUT int	fpConvert;
INPUT int	fpFormat;

{
    PEXPointer	lutBuf;
    int		i;

    switch (tableType)
    {
    case PEXLUTLineBundle:
    {
	pexLineBundleEntry *src;
	PEXLineBundleEntry *dst;

	GetLUTEntryBuffer (numEntries, PEXLineBundleEntry, lutBuf);
	dst = (PEXLineBundleEntry *) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexLineBundleEntry,	pBuf, src);
	    pBuf += SIZEOF (pexLineBundleEntry);

	    dst->type =	src->lineType;
	    dst->interp_method = src->polylineInterp;
	    dst->curve_approx.method = src->curveApprox_method;
	    dst->color.type = src->lineColorType;

	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->curveApprox_tolerance,
		    dst->curve_approx.tolerance, fpFormat);
		FP_CONVERT_NTOH	(src->lineWidth, dst->width, fpFormat);
	    }
	    else
	    {
		dst->curve_approx.tolerance = src->curveApprox_tolerance;
		dst->width = src->lineWidth;
	    }

	    EXTRACT_COLOR_VAL (pBuf, dst->color.type, dst->color.value,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTMarkerBundle:
    {
	pexMarkerBundleEntry *src;
	PEXMarkerBundleEntry *dst;

	GetLUTEntryBuffer (numEntries, PEXMarkerBundleEntry, lutBuf);
	dst = (PEXMarkerBundleEntry *) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexMarkerBundleEntry, pBuf,	src);
	    pBuf += SIZEOF (pexMarkerBundleEntry);

	    dst->type =	src->markerType;
	    dst->color.type = src->markerColorType;
	    
	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->markerScale, dst->scale, fpFormat);
	    }
	    else
		dst->scale = src->markerScale;

	    EXTRACT_COLOR_VAL (pBuf, dst->color.type, dst->color.value,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTTextBundle:
    {
	pexTextBundleEntry *src;
	PEXTextBundleEntry *dst;
	    
	GetLUTEntryBuffer (numEntries, PEXTextBundleEntry, lutBuf);
	dst = (PEXTextBundleEntry *) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexTextBundleEntry,	pBuf, src);
	    pBuf += SIZEOF (pexTextBundleEntry);

	    dst->font_index = src->textFontIndex;
	    dst->precision = src->textPrecision;
	    dst->color.type = src->textColorType;
	    
	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->charExpansion,
		    dst->char_expansion, fpFormat);
		FP_CONVERT_NTOH	(src->charSpacing,
		    dst->char_spacing, fpFormat);
	    }
	    else
	    {
		dst->char_expansion = src->charExpansion;
		dst->char_spacing = src->charSpacing;
	    }

	    EXTRACT_COLOR_VAL (pBuf, dst->color.type, dst->color.value,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTInteriorBundle:
    {
	pexInteriorBundleEntry *src;
	PEXInteriorBundleEntry *dst;

	GetLUTEntryBuffer (numEntries, PEXInteriorBundleEntry, lutBuf);
	dst = (PEXInteriorBundleEntry *) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexInteriorBundleEntry, pBuf, src);
	    pBuf += SIZEOF (pexInteriorBundleEntry);

	    dst->style = src->interiorStyle;
	    dst->style_index = src->interiorStyleIndex;
	    dst->reflection_model = src->reflectionModel;
	    dst->interp_method = src->surfaceInterp;
	    dst->bf_style = src->bfInteriorStyle;
	    dst->bf_style_index	= src->bfInteriorStyleIndex;
	    dst->bf_reflection_model = src->bfReflectionModel;
	    dst->bf_interp_method = src->bfSurfaceInterp;
	    dst->surface_approx.method = src->surfaceApprox_method;

	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->surfaceApproxuTolerance,
		    dst->surface_approx.u_tolerance, fpFormat);
		FP_CONVERT_NTOH	(src->surfaceApproxvTolerance,
		    dst->surface_approx.v_tolerance, fpFormat);
	    }
	    else
	    {
		dst->surface_approx.u_tolerance	= src->surfaceApproxuTolerance;
		dst->surface_approx.v_tolerance	= src->surfaceApproxvTolerance;
	    }

	    /* copy surfaceColor */

	    EXTRACT_COLOR_SPEC (pBuf, dst->color, fpConvert, fpFormat);

	    /* copy reflectionAttr */

	    EXTRACT_REFLECTION_ATTR (pBuf, dst->reflection_attr,
		fpConvert, fpFormat);

	    /* copy bfSurfaceColor */

	    EXTRACT_COLOR_SPEC (pBuf, dst->bf_color, fpConvert,	fpFormat);

	    /* copy bfReflectionAttr */

	    EXTRACT_REFLECTION_ATTR (pBuf, dst->bf_reflection_attr,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTEdgeBundle:
    {
	pexEdgeBundleEntry *src;
	PEXEdgeBundleEntry *dst;
	 
	GetLUTEntryBuffer (numEntries, PEXEdgeBundleEntry, lutBuf);
	dst = (PEXEdgeBundleEntry *) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexEdgeBundleEntry,	pBuf, src);
	    pBuf += SIZEOF (pexEdgeBundleEntry);

	    dst->edge_flag = src->edges;
	    dst->type =	src->edgeType;
	    dst->color.type = src->edgeColorType;

	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->edgeWidth, dst->width, fpFormat);
	    }
	    else
		dst->width = src->edgeWidth;

	    EXTRACT_COLOR_VAL (pBuf, dst->color.type, dst->color.value,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTPattern:
    {
	pexPatternEntry	*src;
	PEXPatternEntry	*dst;
    
	GetLUTEntryBuffer (numEntries, PEXPatternEntry,	lutBuf);
	dst = (PEXPatternEntry *) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexPatternEntry, pBuf, src);
	    pBuf += SIZEOF (pexPatternEntry);

	    dst->color_type = src->colorType;
	    dst->col_count = src->numx;
	    dst->row_count = src->numy;

	    dst->colors.indexed	= (PEXColorIndexed *) PEXAllocBuf (
		GetClientColorSize (dst->color_type) *
		dst->row_count * dst->col_count);

	    EXTRACT_LISTOF_COLOR_VAL (dst->row_count * dst->col_count,
		pBuf, dst->color_type, dst->colors, fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTTextFont:
    {
	pexTextFontEntry *src;
	PEXTextFontEntry *dst;

	GetLUTEntryBuffer (numEntries, PEXTextFontEntry, lutBuf);
	dst = (PEXTextFontEntry	*) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexTextFontEntry, pBuf, src);
	    pBuf += SIZEOF (pexTextFontEntry);

	    dst->count = src->numFonts;

	    dst->fonts = (PEXFont *)
		PEXAllocBuf (dst->count	* sizeof (PEXFont));
	    
	    EXTRACT_LISTOF_CARD32 (dst->count, pBuf, dst->fonts);
	}
	break;
    }

    case PEXLUTColor:
    {
	PEXColorEntry	  *dst;
    
	GetLUTEntryBuffer (numEntries, PEXColorEntry, lutBuf);
	dst = (PEXColorEntry *)	lutBuf;

	EXTRACT_LISTOF_COLOR_SPEC (numEntries, pBuf, dst, fpConvert, fpFormat);
	break;
    }

    case PEXLUTView:
    {
	pexViewEntry *src;
	PEXViewEntry *dst;
	char	     *tPtr;
	 
	GetLUTEntryBuffer (numEntries, PEXViewEntry, lutBuf);
	dst = (PEXViewEntry *) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexViewEntry, pBuf,	src);
	    pBuf += SIZEOF (pexViewEntry);

	    dst->clip_flags = src->clipFlags;

	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->clipLimits_xmin,
		    dst->clip_limits.min.x, fpFormat);
		FP_CONVERT_NTOH	(src->clipLimits_ymin,
		    dst->clip_limits.min.y, fpFormat);
		FP_CONVERT_NTOH	(src->clipLimits_zmin,
		    dst->clip_limits.min.z, fpFormat);
		FP_CONVERT_NTOH	(src->clipLimits_xmax,
		    dst->clip_limits.max.x, fpFormat);
		FP_CONVERT_NTOH	(src->clipLimits_ymax,
		    dst->clip_limits.max.y, fpFormat);
		FP_CONVERT_NTOH	(src->clipLimits_zmax,
		    dst->clip_limits.max.z, fpFormat);
	    }
	    else
	    {
		dst->clip_limits.min.x = src->clipLimits_xmin;
		dst->clip_limits.min.y = src->clipLimits_ymin;
		dst->clip_limits.min.z = src->clipLimits_zmin;
		dst->clip_limits.max.x = src->clipLimits_xmax;
		dst->clip_limits.max.y = src->clipLimits_ymax;
		dst->clip_limits.max.z = src->clipLimits_zmax;
	    }

	    tPtr = (char *) src->orientation;
	    EXTRACT_LISTOF_FLOAT32 (16,	tPtr, dst->orientation,
		fpConvert, fpFormat);

	    tPtr = (char *) src->mapping;
	    EXTRACT_LISTOF_FLOAT32 (16,	tPtr, dst->mapping,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTLight:
    {
	pexLightEntry *src;
	PEXLightEntry *dst;
	
	GetLUTEntryBuffer (numEntries, PEXLightEntry, lutBuf);
	dst = (PEXLightEntry *)	lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexLightEntry, pBuf, src);
	    pBuf += SIZEOF (pexLightEntry);

	    dst->type =	src->lightType;
	    dst->color.type = src->lightColorType;

	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->direction_x, dst->direction.x, fpFormat);
		FP_CONVERT_NTOH	(src->direction_y, dst->direction.y, fpFormat);
		FP_CONVERT_NTOH	(src->direction_z, dst->direction.z, fpFormat);

		FP_CONVERT_NTOH	(src->point_x, dst->point.x, fpFormat);
		FP_CONVERT_NTOH	(src->point_y, dst->point.y, fpFormat);
		FP_CONVERT_NTOH	(src->point_z, dst->point.z, fpFormat);

		FP_CONVERT_NTOH	(src->concentration,
		    dst->concentration,	fpFormat);
		FP_CONVERT_NTOH	(src->spreadAngle,
		    dst->spread_angle, fpFormat);
		FP_CONVERT_NTOH	(src->attenuation1,
		    dst->attenuation1, fpFormat);
		FP_CONVERT_NTOH	(src->attenuation2,
		    dst->attenuation2, fpFormat);
	    }
	    else
	    {
		dst->direction.x = src->direction_x;
		dst->direction.y = src->direction_y;
		dst->direction.z = src->direction_z;

		dst->point.x = src->point_x;
		dst->point.y = src->point_y;
		dst->point.z = src->point_z;

		dst->concentration = src->concentration;
		dst->spread_angle = src->spreadAngle;
		dst->attenuation1 = src->attenuation1;
		dst->attenuation2 = src->attenuation2;
	    }

	    EXTRACT_COLOR_VAL (pBuf, dst->color.type, dst->color.value,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTDepthCue:
    {
	pexDepthCueEntry *src;
	PEXDepthCueEntry *dst;
    
	GetLUTEntryBuffer (numEntries, PEXDepthCueEntry, lutBuf);
	dst = (PEXDepthCueEntry	*) lutBuf;
	
	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexDepthCueEntry, pBuf, src);
	    pBuf += SIZEOF (pexDepthCueEntry);

	    dst->mode =	src->mode;
	    dst->color.type = src->depthCueColorType;

	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->frontPlane, dst->front_plane, fpFormat);
		FP_CONVERT_NTOH	(src->backPlane, dst->back_plane, fpFormat);

		FP_CONVERT_NTOH	(src->frontScaling,
		    dst->front_scaling,	fpFormat);
		FP_CONVERT_NTOH	(src->backScaling,
		    dst->back_scaling, fpFormat);
	    }
	    else
	    {
		dst->front_plane = src->frontPlane;
		dst->back_plane	= src->backPlane;

		dst->front_scaling = src->frontScaling;
		dst->back_scaling = src->backScaling;
	    }

	    EXTRACT_COLOR_VAL (pBuf, dst->color.type, dst->color.value,
		fpConvert, fpFormat);
	}
	break;
    }

    case PEXLUTColorApprox:
    {
	pexColorApproxEntry *src;
	PEXColorApproxEntry *dst;
    
	GetLUTEntryBuffer (numEntries, PEXColorApproxEntry, lutBuf);
	dst = (PEXColorApproxEntry *) lutBuf;
	
	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexColorApproxEntry, pBuf, src);
	    pBuf += SIZEOF (pexColorApproxEntry);

	    dst->type =	src->approxType;
	    dst->model = src->approxModel;
	    dst->max1 =	src->max1;
	    dst->max2 =	src->max2;
	    dst->max3 =	src->max3;
	    dst->dither	= src->dither;
	    dst->mult1 = src->mult1;
	    dst->mult2 = src->mult2;
	    dst->mult3 = src->mult3;
	    dst->base_pixel = src->basePixel;

	    if (fpConvert)
	    {
		FP_CONVERT_NTOH	(src->weight1, dst->weight1, fpFormat);
		FP_CONVERT_NTOH	(src->weight2, dst->weight2, fpFormat);
		FP_CONVERT_NTOH	(src->weight3, dst->weight3, fpFormat);
	    }
	    else
	    {
		dst->weight1 = src->weight1;
		dst->weight2 = src->weight2;
		dst->weight3 = src->weight3;
	    }
	}
	break;
    }

    case (unsigned short) PEXExtLUTTMBinding:
    {
	pexExtTMBindingEntry	*src;
	PEXExtTMBindingEntry	*dst;

	GetLUTEntryBuffer (numEntries, PEXExtTMBindingEntry, lutBuf);
	dst = (PEXExtTMBindingEntry *) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexExtTMBindingEntry, pBuf,	src);
	    pBuf += SIZEOF (pexExtTMBindingEntry);

	    dst->tm_description_id = src->tmDescription; 
	    dst->coord_source_index = src->coordinateLutIndex;
	    dst->composition_index = src->compositionLutIndex;
	    dst->sampling_index	= src->samplingLutIndex;
	}
	break;
    }

    case (unsigned short) PEXExtLUTTMCoordSource: 
    {
	pexExtTMCoordSourceEntry	*src;
	PEXExtTMCoordSourceEntry	*dst;
	char	*tPtr;

	GetLUTEntryBuffer (numEntries, PEXExtTMCoordSourceEntry, lutBuf);
	dst = (PEXExtTMCoordSourceEntry	*) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexExtTMCoordSourceEntry, pBuf, src);
	    pBuf += SIZEOF (pexExtTMCoordSourceEntry);

	    dst->tm_source = src->tmSource; 
	    dst->fp_data_index = src->fpDataIndex;

	    tPtr = (char *) src->orientation;
	    EXTRACT_LISTOF_FLOAT32 (16,	tPtr, dst->orientation,	fpConvert, 
		fpFormat);
	}
	break;
    }

    case (unsigned short) PEXExtLUTTMComposition:
    {
	pexExtTMCompositionEntry	*src;
	PEXExtTMCompositionEntry	*dst;

	GetLUTEntryBuffer (numEntries, PEXExtTMCompositionEntry, lutBuf);
	dst = (PEXExtTMCompositionEntry	*) lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexExtTMCompositionEntry, pBuf, src);
	    pBuf += SIZEOF (pexExtTMCompositionEntry);

	    dst->method	= src->method;	       

	    switch (src->method)
	    {
	    case PEXExtTMCompositeReplace:
	    case PEXExtTMCompositeModulate:
	    case PEXExtTMCompositeDecal:

		break;

	    case PEXExtTMCompositeBlendEnvColor:
	    {
		EXTRACT_CARD16 (pBuf, dst->data.blend_env.channel_selector);
		pBuf +=	2;
		EXTRACT_COLOR_SPEC (pBuf, dst->data.blend_env.color,
		    fpConvert, fpFormat);
		break;
	    }

	    case PEXExtTMCompositeDecalBackground:
	    {
		EXTRACT_COLOR_SPEC (pBuf, dst->data.decal_bkgd_color,
		    fpConvert, fpFormat);
		break;
	    }

	    case PEXExtTMCompositeReplaceBlendedColors:
	    {
		EXTRACT_CARD16 (pBuf, dst->data.blend_repl.channel_selector);
		pBuf +=	2;
		EXTRACT_COLOR_SPEC (pBuf, dst->data.blend_repl.color1,
		    fpConvert, fpFormat);
		EXTRACT_COLOR_SPEC (pBuf, dst->data.blend_repl.color2,
		    fpConvert, fpFormat);
		break;
	    }

	    default:
	    {
		EXTRACT_CARD16 (pBuf, dst->data.imp_dep.length);
		dst->data.imp_dep.imp_dep = (char *)
		    PEXAllocBuf	(dst->data.imp_dep.length);

		COPY_AREA ((char *) pBuf, dst->data.imp_dep.imp_dep, 
		    dst->data.imp_dep.length);
		pBuf +=	dst->data.imp_dep.length;
		break;
	    }
	    }
	}
	break;
    }

    case (unsigned short) PEXExtLUTTMSampling: 
    {
	pexExtTMSamplingEntry	*src;
	PEXExtTMSamplingEntry	*dst;
	char	*tPtr;

	GetLUTEntryBuffer (numEntries, PEXExtTMSamplingEntry, lutBuf);
	dst = (PEXExtTMSamplingEntry *)	lutBuf;

	for (i = 0; i <	numEntries; i++, dst++)
	{
	    GET_STRUCT_PTR (pexExtTMSamplingEntry, pBuf, src);
	    pBuf += SIZEOF (pexExtTMSamplingEntry);

	    dst->minification_method = src->minification;
	    dst->magnification_method =	src->magnification;
	    dst->t0_boundary_condition = src->t0Boundary;
	    dst->t1_boundary_condition = src->t1Boundary;
	    dst->t2_boundary_condition = src->t2Boundary;
	    dst->boundary_clamp_color_source = src->clampColorSource;

	    EXTRACT_COLOR_SPEC (pBuf, dst->clamp_color,	fpConvert, fpFormat);

	    EXTRACT_FLOAT32 (pBuf, dst->depth_sampling_bias_hint,
		fpConvert, fpFormat);

	    EXTRACT_FLOAT32 (pBuf, dst->t0_frequency_hint, fpConvert, fpFormat);
	    EXTRACT_FLOAT32 (pBuf, dst->t1_frequency_hint, fpConvert, fpFormat);
	    EXTRACT_FLOAT32 (pBuf, dst->t2_frequency_hint, fpConvert, fpFormat);
	}
	break;
    }
    }

    return (lutBuf);
}
/***** End of Ext section ************/
